import datetime
import os
import sys
from pathlib import Path

# Ensure documentation examples are deterministically random.
import numpy
import tomli
from pkg_resources import get_distribution

try:
    numpy.random.seed(int(os.environ["SOURCE_DATE_EPOCH"]))
except KeyError:
    pass

try:
    from sphinx_astropy.conf.v1 import *  # noqa
except ImportError:
    print("ERROR: the documentation requires the sphinx-astropy package to be installed")
    sys.exit(1)

# Get configuration information from `pyproject.toml`
with open(Path(__file__).parent.parent / "pyproject.toml", "rb") as configuration_file:
    conf = tomli.load(configuration_file)
configuration = conf["project"]

# -- General configuration ----------------------------------------------------

project = configuration["name"]
author = configuration["authors"][0]["name"]
copyright = f"{datetime.datetime.now().year}, {author}"

release = get_distribution(configuration["name"]).version
# The short X.Y version
version = ".".join(release.split(".")[:2])

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.2'

intersphinx_mapping["pypa-packaging"] = ("https://packaging.python.org/en/latest/", None)  # noqa
intersphinx_mapping["asdf-astropy"] = ("https://asdf-astropy.readthedocs.io/en/latest/", None)  # noqa
intersphinx_mapping["pytest"] = ("https://docs.pytest.org/en/latest/", None)  # noqa

subprojects = {
    # main project
    "asdf-website": ("https://www.asdf-format.org/en/latest", None),
    # other subprojects
    "asdf": ("https://www.asdf-format.org/projects/asdf/en/stable/", None),
    "asdf-standard": ("https://asdf-standard.readthedocs.io/en/latest/", None),
    "asdf-transform-schemas": ("https://www.asdf-format.org/projects/asdf-transform-schemas/en/latest/", None),
    "asdf-wcs-schemas": ("https://www.asdf-format.org/projects/asdf-wcs-schemas/en/latest/", None),
}

intersphinx_mapping.update(subprojects)  # noqa

# Adds a global navigation in the topbar - consistent across subprojects
globalnavlinks = {
    "ASDF Projects": "https://www.asdf-format.org",
    "Tutorials": "https://www.asdf-format.org/en/latest/tutorials/index.html",
    "Community": "https://www.asdf-format.org/en/latest/community/index.html",
}
topbanner = ""
for text, link in globalnavlinks.items():
    topbanner += f"<a href={link}>{text}</a>"

# To perform a Sphinx version check that needs to be more specific than
# major.minor, call `check_sphinx_version("x.y.z")` here.
# check_sphinx_version("1.2.1")

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns.append("_templates")  # noqa

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog += """"""  # noqa


# -- Options for HTML output ---------------------------------------------------

# A NOTE ON HTML THEMES
# The global astropy configuration uses a custom theme, 'bootstrap-astropy',
# which is installed along with astropy. A different theme can be used or
# the options for this theme can be modified by overriding some of the
# variables set in the global configuration. The variables set in the
# global configuration are listed below, commented out.

# Add any paths that contain custom themes here, relative to this directory.
# To use a different custom theme, add the directory containing the theme.
# html_theme_path = []

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes. To override the custom theme, set this to the
# name of a builtin theme or the name of a custom theme in html_theme_path.
html_theme = "furo"

html_static_path = ["_static"]

# Custom sidebar templates, maps document names to template names.
# Override default settings from sphinx_asdf / sphinx_astropy (incompatible with furo)
html_sidebars = {}

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/images/favicon.ico"
html_logo = ""
html_theme_options = {
    "light_logo": "images/logo-light-mode.png",
    "dark_logo": "images/logo-dark-mode.png",
    "announcement": topbanner,
}

pygments_style = "monokai"

# NB Dark style pygments is furo-specific at this time
pygments_dark_style = "monokai"

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = ''

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = f"{project.replace('_', ' ')} v{release}"

# Output file base name for HTML help builder.
htmlhelp_basename = project + "doc"

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    "-Nfontsize=10",
    "-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Efontsize=10",
    "-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Gbgcolor=white",
    "-Gfontsize=10",
    "-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
]

# -- Options for LaTeX output --------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [("index", project + ".tex", project + " Documentation", author, "manual")]

latex_logo = "_static/images/logo-light-mode.png"


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", project.lower(), project + " Documentation", [author], 1)]

sys.path.insert(0, os.path.join(os.path.abspath(os.path.dirname("__file__")), "sphinxext"))
extensions += ["sphinx_asdf", "sphinx.ext.intersphinx", "sphinx.ext.extlinks"]  # noqa


# -- sphinx_asdf configuration ---------------------------------------------

# Top-level directory containing ASDF schemas (relative to current directory)
asdf_schema_path = "../resources"
# This is the prefix common to all schema IDs in this repository
asdf_schema_standard_prefix = "schemas"


def setup(app):
    app.add_css_file("custom.css")



import re

import asdf
import pytest

ALLOWED_REFS = (
    r"^baseframe-[0-9.]+$",
    r"^frames/baseframe-[0-9.]+$",
    r"^#.*$",
)

UNIT_TAGS = {"tag:stsci.edu:asdf/unit/unit-1.*", "tag:astropy.org:astropy/units/unit-1.*"}


def test_only_known_refs(latest_schema):
    """Latest schemas should only contain specific refs"""
    for node in asdf.treeutil.iter_tree(latest_schema):
        if not isinstance(node, dict):
            continue
        if "$ref" in node:
            uri = node["$ref"]
            if not any(re.match(pattern, uri) for pattern in ALLOWED_REFS):
                assert False, f"Unexpected $ref: {uri}"


def test_wildcard_tags(latest_schema):
    """Latest schemas should only contain wildcarded tags"""
    for node in asdf.treeutil.iter_tree(latest_schema):
        if not isinstance(node, dict):
            continue
        if "tag" in node:
            pattern = node["tag"]
            if "*" not in pattern:
                assert False, f"tag pattern missing wildcard: {pattern}"


@pytest.mark.parametrize("tag_set", (UNIT_TAGS,))
def test_tags_in_allof(latest_schema, tag_set):
    """
    Test that some tags (unit) where the
    tag used depends on the value are always referenced in an
    allof containing all tags.
    """
    # we walk_and_modify here to use postorder

    def callback(node):
        if not isinstance(node, dict):
            return node
        if "anyOf" in node:
            # check if this anyof includes a tag in the set
            seen = set()
            for sub in node["anyOf"]:
                if not isinstance(sub, dict):
                    continue
                if "tag" not in sub:
                    continue
                if sub["tag"] in tag_set:
                    seen.add(sub["tag"])

            if seen:
                # if a tag was found, check both were found
                assert seen == tag_set, f"anyOf {node} missing: {tag_set - seen}"
                # remove the anyof so the code below can check for tags not in the anyof
                return {k: v for k, v in node.items() if k != "anyOf"}
        if tag := node.get("tag"):
            assert tag not in tag_set, f"tag {tag} must be in an anyOf with: {tag_set}"
        return node

    asdf.treeutil.walk_and_modify(latest_schema, callback, postorder=False)



from pathlib import Path

import asdf
import pytest
import yaml


def get_latest_schema_uris(resource_paths):
    by_base_uri = {}

    for resource_path in resource_paths:
        # skip manifests
        if "manifests" in resource_path.parts:
            continue

        # get uri from schema id
        with resource_path.open("rb") as f:
            schema_uri = yaml.safe_load(f.read())["id"]

        # get base uri and version
        base, version = schema_uri.rsplit("-", maxsplit=1)
        if base not in by_base_uri:
            by_base_uri[base] = {}
        by_base_uri[base][version] = schema_uri
    return [by_version[max(by_version.keys())] for _, by_version in by_base_uri.items()]


RESOURCE_PATHS = list((Path(__file__).parent.parent / "resources").glob("**/*.yaml"))
MANIFEST_PATHS = [p for p in RESOURCE_PATHS if "manifests" in p.parts]
SCHEMA_PATHS = [p for p in RESOURCE_PATHS if "schemas" in p.parts]
LATEST_SCHEMA_URIS = get_latest_schema_uris(RESOURCE_PATHS)


@pytest.fixture(params=RESOURCE_PATHS)
def resource_path(request):
    yield request.param


@pytest.fixture(params=SCHEMA_PATHS)
def schema_path(request):
    yield request.param


@pytest.fixture(params=MANIFEST_PATHS)
def manifest_path(request):
    yield request.param


@pytest.fixture
def latest_manifest_path():
    yield max(MANIFEST_PATHS)


@pytest.fixture
def latest_manifest(latest_manifest_path):
    with latest_manifest_path.open("rb") as f:
        yield yaml.safe_load(f.read())


@pytest.fixture
def latest_schema_uris():
    yield LATEST_SCHEMA_URIS


@pytest.fixture(params=LATEST_SCHEMA_URIS)
def latest_schema_uri(request):
    yield request.param


@pytest.fixture()
def latest_schema(latest_schema_uri):
    yield asdf.schema.load_schema(latest_schema_uri)



"""
Test that the asdf library integration is working properly.
"""

import asdf
import yaml


def test_manifest_integration(manifest_path):
    content = manifest_path.read_bytes()
    manifest = yaml.safe_load(content)
    asdf_content = asdf.get_config().resource_manager[manifest["id"]]
    assert asdf_content == content


def test_schema_integration(schema_path):
    content = schema_path.read_bytes()
    schema = yaml.safe_load(content)
    asdf_content = asdf.get_config().resource_manager[schema["id"]]
    assert asdf_content == content



"""
Test that the manifest file is correctly structured and refers
to schemas that exist.
"""

import asdf
import yaml


def test_manifest(manifest_path):
    content = manifest_path.read_bytes()
    manifest = yaml.safe_load(content)
    schema = asdf.schema.load_schema("asdf://asdf-format.org/core/schemas/extension_manifest-1.0.0")

    asdf.schema.validate(manifest, schema=schema)

    for tag in manifest.get("tags", []):
        # Check that the schema exists:
        if "schema_uri" in tag:
            assert tag["schema_uri"] in asdf.get_config().resource_manager


def test_manifest_tag_order(latest_manifest):
    """Tags should be sorted alphabetically"""
    tag_uris = [tag_def["tag_uri"] for tag_def in latest_manifest["tags"]]
    assert tag_uris == sorted(tag_uris)


def test_tags_match_schemas(latest_manifest):
    """Check that tag and schema versions match"""
    for tag_def in latest_manifest["tags"]:
        tag_uri = tag_def["tag_uri"]
        schema_uri = tag_def["schema_uri"]
        assert tag_uri.split(":")[-1] in schema_uri


def test_uses_latest_schemas(latest_manifest, latest_schema_uris):
    """The latest manifest should always use the latest schemas"""
    for tag_def in latest_manifest["tags"]:
        schema_uri = tag_def["schema_uri"]
        assert schema_uri in latest_schema_uris



import importlib.resources as importlib_resources
from pathlib import Path

from asdf.resource import DirectoryResourceMapping

import asdf_coordinates_schemas


def get_resource_mappings():
    resources_root = importlib_resources.files(asdf_coordinates_schemas) / "resources"
    if not resources_root.is_dir():
        # In an editable install, the resources directory will exist off the
        # repository root:
        resources_root = Path(__file__).absolute().parent.parent.parent / "resources"
        if not resources_root.is_dir():
            raise RuntimeError("Missing resources directory")

    return [
        DirectoryResourceMapping(
            resources_root / "schemas",
            "http://astropy.org/schemas/astropy/coordinates/",
            recursive=True,
        ),
        DirectoryResourceMapping(
            resources_root / "manifests",
            "asdf://asdf-format.org/astronomy/coordinates/manifests/",
        ),
    ]
